#
#  engine
#
#  Playout Daemon for autoradio project
#
#
#  Copyright (C) 2017-2018 Gottfried Gaisbauer <gottfried.gaisbauer@servus.at>
#
#  This file is part of engine.
#
#  engine is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  any later version.
#
#  engine is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with engine. If not, see <http://www.gnu.org/licenses/>.
#

import logging

from libraries.base.config import AuraConfig


class AuraLogger():
    config = None
    logger = None

    def __init__(self, config):
        self.config = config

        self.__create_logger("AuraEngine")

    def __create_logger(self, name):
        """
        Creates the logger instance for AuraEngine
        :param name: LoggerName
        :return:
        """
        lvl = self.config.get("loglevel")

        # create logger
        self.logger = logging.getLogger(name)
        self.logger.setLevel(lvl)

        if not self.logger.hasHandlers():
            # create file handler for logger
            file_handler = logging.FileHandler(self.config.get("logdir") + "/engine.log")
            file_handler.setLevel(lvl)

            # create stream handler for logger
            stream_handler = logging.StreamHandler()
            stream_handler.setLevel(lvl)

            # set format of log
            datepart = "%(asctime)s:%(name)s:%(levelname)s"
            message = " - %(message)s - "
            filepart = "[%(filename)s:%(lineno)s-%(funcName)s()]"
            formatter = logging.Formatter(datepart + message + filepart)

            # set log of handlers
            file_handler.setFormatter(formatter)
            stream_handler.setFormatter(formatter)

            # add handlers to the logger
            self.logger.addHandler(file_handler)
            self.logger.addHandler(stream_handler)

            self.logger.critical("ADDED HANDLERS")
        else:
            self.logger.critical("REUSED LOGGER")